// Microsoft SharePoint connector's configuration settings
const config = {
    credential: {
		// Please specify following Microsoft SharePoint credentials
        // Client Id guid
        client_id: "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
		// Client secret string
        client_secret: "xxxxxxxxxxxxxxxxxx",
		// Tenant guid
        tenant: "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"
    },
    export: {
		//Microsoft SharePoint site link
		sharepoint_site_link: "https://yourInstance.sharepoint.com/YourSite/YourLibrary/Forms/AllItems.aspx",

        //ABBYY Vantage exports a wide range of formats: PDF, JSON, CUSTOM_JSON,  TEXT, DOCX, XLSX, FIELDS_JSON, OCR_JSON
        //With this filter you can select which formats need to be exported to Microsoft SharePoint.
        //Multiple selections are possible.
        //For PDF and Text, ABBYY Vantage will automatically export metadata as well according to your Content type & corresponding Mapping.
        export_format_filter: "PDF, JSON, CUSTOM_JSON, TEXT, DOCX, XLSX, FIELDS_JSON, OCR_JSON",

        //Microsoft SharePoint supports multiple content types, set the name of the content type to export the document to.
        //If this is blank, the default content type will be used.
        content_type: "Invoice",

        //ABBYY Vantage will only export the files that correspond to the filter below to the content type specified in 'content_type'
        //Other files defined in the 'export_format_filter' variable will be exported as the default content type.
        //Possible values are: PDF, JSON, CUSTOM_JSON, TEXT, DOCX, XLSX, FIELDS_JSON, OCR_JSON, ""
        //If this variable is empty, PDF and TEXT formats will be automatically exported to the configured content type.
        content_type_export_filter: "PDF",

        //This variable allows you to filter which export formats defined in the 'export_format_filter' variable should also have the metadata updated in Microsoft SharePoint.
        //If this variable is empty, PDF and TEXT exports will have metadata transfer enabled by default.
        //PDF, JSON, CUSTOM_JSON, TEXT, DOCX, XLSX, FIELDS_JSON, OCR_JSON, ""
        vantage_fields_to_sharepoint_metadata_filter: "", 

        //Choose whether a record should be created in a Microsoft SharePoint list or a new file should be uploaded.
        //The default value is 'UPLOAD_FILE'
        //CREATE_LIST_RECORD, UPLOAD_FILE
        export_scenario: "UPLOAD_FILE",

        //This variable allows you to change the behavior of the connector when a file already exists.
        //DO_NOT_UPLOAD_IF_FILE_EXISTS, INCREASE_VERSION, CREATE_COPY, OVERWRITE
        file_uploading_policy: "INCREASE_VERSION",

        //$(ExtractedData.FieldName)- the variable will be replaced with the value of the extracted data field
        //$(FullFileName) - the variable will be replaced with original Vantage document file name 
        //$(Date) - The current date yyyy-MM-dd. 
		//$(Time) - The current time HH-mm-ss. 
		//$(DateTime) - The current date time yyyy-MM-dd HH-mm-ss. 
        //$(TransactionID) - The identifier of the transaction. 
        file_name_schema: "$(FullFileName)",
		
    },
    mapping: [
        {
            field: {
                sharepoint_field_name: "InvoiceNumber",
                vantage_field_name: "Invoice Number"
            }

        }
    ]
}

// JsonCreator module creates simple json from the original extracted data document and selected mapping 
let exportedDocs = Context.Transaction.Documents.length;
let fieldsMapped = false;
const JsonCreator = (function () {
    const creator = {};

    creator.GenerateCustomJson = function (document, mapping) {
        let jsonObj = {};
        for (let i = 0; i < mapping.length; i++) {
            let value;
            let fieldName = mapping[i].field.vantage_field_name;
            if (fieldName) {
                let documentField = document.GetField(fieldName);
				
				if (documentField.DataType === DataType.Amount)
					value = documentField.Value;
				else if (documentField.DataType === DataType.Date)
					value = new Date(documentField.Value).toISOString();
				else if (documentField.DataType === DataType.None)
				{
					if (documentField.Instances.length > 0)
					{
						let jsonArray = new Array();
						for (let instance of documentField.Instances) 
						{
							let jsonObject = {};
							for (let child of instance.Children)
							{
								
								let cValue = "";
								if (child.Value)
									cValue = child.Value.toString();
								jsonObject[child.Name] = cValue;
							}
							jsonArray.push(jsonObject);
						}
						value = jsonArray;
					}
					else if (documentField.Children)
					{
						let jsonObject = {};
						for (let child of documentField.Children)
						{
							let cValue = "";
							if (child.Value)
								cValue = child.Value.toString();
							jsonObject[child.Name] = cValue;
						}
						value = jsonObject;
					}
				}
				else if (documentField.Instances.length > 0)
				{
					let arr = [];
					for (let instance of documentField.Instances)
						arr.push(instance.Value);
					value = arr;
				}
				else
					value = documentField.Value;
				
            } else {
                fieldName = mapping[i].field.sharepoint_field_name;
                value = mapping[i].field.value;
            }

            jsonObj[fieldName] = value;
        }
        return JSON.stringify(jsonObj);
    };

    return creator;
}());

// SharePoint connector's main module
const SharePointExporter = (function ($config) {
    const exporter = {};

    const client_id = $config.credential.client_id;
    const client_secret = $config.credential.client_secret;
    const tenant = $config.credential.tenant;
    const sharePointUrl = $config.export.sharepoint_site_link;
	let export_scenario = 'UPLOAD_FILE';
	if ($config.export.export_scenario)
		export_scenario = $config.export.export_scenario.toUpperCase();
    const file_name_schema = $config.export.file_name_schema;
	let export_format = 'PDF, TEXT';
	let defaultValues = false;
	if ($config.export.export_format_filter)
		export_format = $config.export.export_format_filter;
	else
		defaultValues = true;
    let export_format_array = export_format.split(',');
	removeSpacesFromArray(export_format_array);
	let eport_format_obj_array = [];
	for (let el of export_format_array)
		eport_format_obj_array.push({});
	
	let formats_for_fields = $config.export.vantage_fields_to_sharepoint_metadata_filter;
    let formats_for_fields_array = formats_for_fields.split(',');
    const mapping = $config.mapping;
	let contentType = $config.export.content_type;
	let content_type_formats = $config.export.content_type_export_filter;
	
    let content_type_formats_array = content_type_formats.split(',');
    const file_uploading_policy = $config.export.file_uploading_policy.toUpperCase();

    const apiDomain = "https://graph.microsoft.com/v1.0";
    const apiSitesUrl = 'https://graph.microsoft.com/v1.0/sites/';

    //getting Bearer token
    const token = getToken();
    if (!token)
        throw "The SharePoint token is missing";
    //getting siteLink by parsing the sharePointUrl
    const siteLink = getSiteLink(sharePointUrl);
	//getting siteObject that contains site id and site webUrl
    const siteObject = getSiteObject(siteLink);
	
	let fileObject = {
		token: null,
		driveId: null,
		rootFolder: null,
		fileName: null, 
		exportResult: null,
		contentTypeId: null,
		drivePath: null,
		contentTypeName: null,
		changesJson: {},
		hasErrors: false,
		exportFormat: null,
		listId: null,
		contentTypeLink: null,
		existsInitially: false
	}
	
	Context.LogMessage(`Connector is about to export the following formats: ${export_format_array.join(", ")}`);
	// document - one of Context.Transaction.Documents[]
    exporter.Export = function (document, counter) {
		
		try
		{
			Context.LogMessage(`Starting to export document, using scenario: ${export_scenario}, file naming schema ${file_name_schema}`);
			
			if (export_scenario === 'UPLOAD_FILE') {
				fileObject.listId = getListId();
				if (fileObject.listId)
					fileObject.contentTypeLink = `${apiSitesUrl}${siteObject.id}/lists/${fileObject.listId}/contentTypes`
				else
					fileObject.contentTypeLink = `${apiSitesUrl}${siteObject.id}/contentTypes`

				//getting drive Url
				const driveUrl = getDriveUrl(sharePointUrl, siteObject.webUrl);
				//getting drive id by site id
				const driveId = getDriveIdBySiteId(siteObject.id, driveUrl);
				//getting folder path on the disk by parsing the sharePointUrl
				const rootFolder = getRootFolder(sharePointUrl, driveUrl);
				const fileName = getFileName(document, file_name_schema);
				

				const exports = document.Exports;
				if (defaultValues)
				{
					if (!content_type_formats_array[0])
						content_type_formats_array = export_format_array;
					if (!formats_for_fields_array[0])
						formats_for_fields_array = export_format_array;
				}
				removeSpacesFromArray(formats_for_fields_array);
				removeSpacesFromArray(content_type_formats_array);
				for (let i = 0; i < export_format_array.length; i++) {
					
					clearFileObject();
					Context.LogMessage(`Starting to upload: '${export_format_array[i]}' file`);
					//detect needed format, upload file and write metadate if needed
					const exportFormat = export_format_array[i].replace(' ', '').toUpperCase();
					Context.LogMessage(`Export format: ${exportFormat},  file uploading policy: ${file_uploading_policy}`);
					let exportFormatForFileName = '.' + exportFormat.toLowerCase();
					if (exportFormat === 'CUSTOM_JSON')
						exportFormatForFileName = ' CUSTOM.json';
					else if (exportFormat === 'FIELDS_JSON')
						exportFormatForFileName = ' FIELDS.json';
					else if (exportFormat === 'OCR_JSON')
						exportFormatForFileName = ' OCR.json';
					else if (exportFormat === 'TEXT')
						exportFormatForFileName = '.txt';

					
					let exportResult;
					switch(exportFormat) {
						case 'PDF':
							exportResult = getExportResult(exports, ExportFormat.Pdf);
							break;
						case 'JSON':
							exportResult = getExportResult(exports, ExportFormat.Json);
							break;
						case 'TEXT':
							exportResult = getExportResult(exports, ExportFormat.Text);
							break;
						case 'DOCX':
							exportResult = getExportResult(exports, ExportFormat.Docx);
							break;
						case 'XLSX':
							exportResult = getExportResult(exports, ExportFormat.Xlsx);
							break;
						case 'FIELDS_JSON':
							exportResult = getExportResult(exports, ExportFormat.FieldsJson);
							break;
						case 'OCR_JSON':
							exportResult = getExportResult(exports, ExportFormat.OcrJson);
							break;
						case 'CUSTOM_JSON':
							if (mapping.length === 0) {
								Context.LogMessage("No fields in mapping found, skipping the CUSTOM_JSON exporting");
								continue;
							}
							exportResult = JsonCreator.GenerateCustomJson(document, mapping);
							break;
						default: throw `Invalid export format: ${exportFormat}`; 
					}
					if (!exportResult) {
						Context.LogMessage(`Export result is missing in the exports' collection of Vantage document for the format '${exportFormat}'. Please enable this format exporting in the Activity Exported Data > Settings. Skipping this file...`);
						eport_format_obj_array[i]['failed'] = true;
						continue;
					}
					else{
						fileObject.token = token;
						fileObject.driveId = driveId;
						fileObject.rootFolder = rootFolder;
						fileObject.fileName = fileName + exportFormatForFileName;
						fileObject.exportResult = exportResult;
						fileObject.exportFormat = exportFormat;
						fileObject.drivePath = 	`/drives/${fileObject.driveId}/root:${fileObject.rootFolder}`;
						fileObject.existsInitially = checkIfFileExists();
						//Creating a JSON with content type and fields
						
						if (fileObject.existsInitially && file_uploading_policy === 'DO_NOT_UPLOAD_IF_FILE_EXISTS')
						{
							Context.LogMessage(`WARNING: File ${fileObject.fileName} already exists, skipping it`);
							continue;
						}
						if (fileObject.existsInitially && file_uploading_policy === 'OVERWRITE')
						{
							getChangesJson(document, counter, i);
							if (performeChanges(false).hasErrors)
								throw `File was not uploaded. Could not apply fields`;
						}
						
						fileObject.drivePath = uploadFile();
						if (fileObject.drivePath == 409)
							continue;
						
						getChangesJson(document, counter, i);
					}
					
					performeChanges(true);
					Context.LogMessage(`File '${export_format_array[i]}' has been successfully uploaded`);
					eport_format_obj_array[i]['failed'] = false;
				}
				let fullFailed = true;
				for (let el of eport_format_obj_array)
				{
					if (!el.failed)
						fullFailed = false;
				}
				if (fullFailed)
				{
					Context.LogMessage(`Document ${counter} has not been exported`);
					exportedDocs--;
				}
				else
					Context.LogMessage(`Document ${counter} has been successfully exported`);
						
			} else if (export_scenario === 'CREATE_LIST_RECORD') {
				//cheking if the mapping object contains any fields
				if (mapping.length > 0) {
					//getting list id
					fileObject.listId = getListId();
					fileObject.contentTypeLink = `${apiSitesUrl}${siteObject.id}/lists/${fileObject.listId}/contentTypes`;
					getListJson(document, counter);
					createListRecord(document);
				} else 
					throw `No fields where configured for the ${export_scenario}`;
			}
		}
		catch(Exception)
		{
			if (!fileObject.existsInitially && export_scenario === 'UPLOAD_FILE')
				deleteFile(true, Exception);
			else
				throw Exception;
		}
    };
	
	function clearFileObject()
	{
		const keys = Object.keys(fileObject);
		for (let key of keys)
        if (key === 'hasErrors' || key === 'existsInitially')
            fileObject[key] = false;
        else if (key === 'changesJson')
            fileObject[key] = {};
		else if (key === 'listId' || key === 'contentTypeLink')
			continue
        else
            fileObject[key] = null;
	}
	
	function getContentTypeId(cpIndex)
	{
		if (!contentType)
			Context.LogMessage(`WARNING! Content type is empty. Setting the default one`);
		let contentTypeId = '';
		if (cpIndex >=0 && contentType) 
			contentTypeId = getContentSpecifiedTypeId(contentType);
		else
			contentTypeId = getContentTypeOfExistingFile();
		fileObject.contentTypeId = contentTypeId;
		if (!fileObject.contentTypeName)
			fileObject.contentTypeName = contentType;
	}
	
	function getChangesJson(document, docCounter, formatCounter)
	{
		const cpIndex = content_type_formats_array.findIndex(element => {
			return replaceAll(' ', '', element.toUpperCase()) === fileObject.exportFormat.toUpperCase();
		});
		const fIndex = formats_for_fields_array.findIndex(element => {
			return replaceAll(' ', '', element.toUpperCase()) === fileObject.exportFormat.toUpperCase();
		});
		
		getContentTypeId(cpIndex);
		
	
		if (!fieldsMapped)
			changeColumnsNameAccordingToSPNamesFile();
		
		if (cpIndex >= 0)
			fileObject.changesJson["contentType"] = createContentTypeJson();
		if (fIndex >= 0)
			fileObject.changesJson["fields"] = createFieldsJson(document);
	}
	
	function getListJson(document, docCounter)
	{
		let contentTypeId = '';
		if (contentType) 
			contentTypeId = getContentSpecifiedTypeId(contentType);
		
		fileObject.contentTypeId = contentTypeId;
		fileObject.contentTypeName = contentType;
		
		if (docCounter === 1)
			changeColumnsNameAccordingToSPNamesList();
		
		fileObject.changesJson["fields"] = createFieldsJson(document);
		if (contentType)
			fileObject.changesJson["contentType"] = createContentTypeJson();
		
	}
	
	function performeChanges(startLog)
	{
		const json = fileObject.changesJson;
		if (!json.contentType && !json.fields)
			return;
		if (startLog)
		{
			if (!json.contentType)
				Context.LogMessage(`Performing changes to the document "${fileObject.fileName}"; Setting fields`);
			else if (!json.fields)
				Context.LogMessage(`Performing changes to the document "${fileObject.fileName}"; Setting content type`);
			else
				Context.LogMessage(`Performing changes to the document "${fileObject.fileName}"; Setting content type along with fields`);
		}
		
		const request = Context.CreateHttpRequest();
		request.ThrowExceptionOnFailed = false;
		request.Url = `${apiDomain}${fileObject.drivePath}/${fileObject.fileName}:/listItem`;
		request.Method = 'PATCH';
		request.SetStringContent(JSON.stringify(json), ContentEncoding.UTF8, "application/json");
		request.AuthToken = token;
		request.AuthScheme = 'Bearer';
		request.Send();
		const jsonResponse = JSON.parse(request.ResponseText);
		if (request.Status < 200 || request.Status > 299){
			Context.LogMessage(`Failed performing changes to the document "${fileObject.fileName}"`);
			if (file_uploading_policy === 'OVERWRITE')
			{
				Context.LogMessage(`SharePoint Error: ${jsonResponse.error.message}`);
				fileObject.hasErrors = true;
				return fileObject;
			}				
			deleteFile(true, `Fields setting failed; SharePoint error: ${jsonResponse.error.message}`);
		}
		else
		{
			if (!json.contentType)
				Context.LogMessage(`All fields have been successfully set to the document "${fileObject.fileName}"`);
			else if (!json.fields)
				Context.LogMessage(`Content type ${fileObject.contentTypeName} has been successfully set to the document "${fileObject.fileName}"`);
			else
				Context.LogMessage(`Content type ${fileObject.contentTypeName} and all fields have been successfully set to the document "${fileObject.fileName}"`);
		}
			
	}
	function createContentTypeJson()
	{
		let json = {};
		json["id"] = fileObject.contentTypeId;
		return json;
	}
	function getExportResult(collection, format) {
		try {
			return collection.find(element => element.ExportFormat === format);
		} catch {
		}
	}
	function getContentTypeOfExistingFile(){
		const request = Context.CreateHttpRequest();
		request.ThrowExceptionOnFailed  = true;
        request.Url = apiDomain + fileObject.drivePath + '/' + fileObject.fileName + ':/listItem';
        request.Method = 'GET';
        request.AuthToken = token;
        request.AuthScheme = 'Bearer';
        request.Send();
		
		let jsonResponse = JSON.parse(request.ResponseText);
		if (request.Status == 200) 
		{
			fileObject.contentTypeName = jsonResponse.contentType.name;
			return jsonResponse.contentType.id;
		}
		else
			throw `SharePoint error: ${jsonResponse.error.message}`;
	}
    function getContentSpecifiedTypeId(contentType) {
        const request = Context.CreateHttpRequest();
        request.Url = fileObject.contentTypeLink;
        request.Method = 'GET';
        request.AuthToken = token;
        request.AuthScheme = 'Bearer';
        request.Send();

        const jsonResponse = JSON.parse(request.ResponseText);
        if (request.Status == 200) {
			const values = jsonResponse.value;
			for (let value of values) {
				if (value.name === contentType)
					return value.id;
			}
			throw `Specified content type '${contentType}' has not been found`;
		}
		throw `SharePoint error: ${jsonResponse.error.message}`;
    }

    function getListWebUrl() {
        //parsing link to identify the list web url
        if (sharePointUrl.includes('/:l:/r')) 
			return sharePointUrl.replace('/:l:/r', '');
		else if (sharePointUrl.includes('/:f:/r')) 
			return sharePointUrl.replace('/:f:/r', '');
        else {
            const parts = sharePointUrl.split('/');
            let tempSharePointUrl = '';
			let needToStop = false;
            for (let i = 0; i < parts.length - 1; i++) {
				if (export_scenario === 'UPLOAD_FILE')
				{
					tempSharePointUrl += parts[i]
					if (needToStop)
						break;
					if (tempSharePointUrl === siteObject.webUrl)
						needToStop = true;
					tempSharePointUrl += '/';
				}
				else
				{
					if (i === parts.length - 2)
						tempSharePointUrl += parts[i];
					else
						tempSharePointUrl += parts[i] + '/';
				}
				
            }
            return tempSharePointUrl;
        }
    }

    function getListId(nextLink, foundValue) {
		if (!foundValue)
			foundValue = {webUrl: '', id: ''};
        let listWebUrl = getListWebUrl();
        const request = Context.CreateHttpRequest();
        if (nextLink)
            request.Url = nextLink;
        else
            request.Url = apiSitesUrl + siteObject.id + '/lists';
        request.Method = 'GET';
        request.AuthToken = token;
        request.AuthScheme = 'Bearer';
        request.Send();

        const jsonResponse = JSON.parse(request.ResponseText);
		if (request.Status == 200) {
			//trying to get '@odata.nextLink' property that contains URL for the next request
			//if this responce returns more than 200 values
			const propertyName = '@odata.nextLink';
			const nextLink = jsonResponse[propertyName];
			const values = jsonResponse.value;
			for (let value of values) {
				if (value.webUrl === listWebUrl) {
					return value.id;
				}
				else if (listWebUrl.includes(value.webUrl) && value.webUrl.length > foundValue.webUrl.length)
					foundValue = value;
			}
			//call function again passing nextLink if needed value has not been found
			//and the nextLink property is not empty
			if (nextLink) 
				getListId(nextLink);
			if (foundValue)
				return foundValue.id;
			if (export_scenario === 'CREATE_LIST_RECORD')
				throw `The specified SharePoint list (${listWebUrl}) has not been found`;
		}
        throw `SharePoint error ${jsonResponse.error.message}`;
    }
	
	function changeColumnsNameAccordingToSPNamesFile() {
		
		
		const request = Context.CreateHttpRequest();
        request.Url = `${fileObject.contentTypeLink}/${fileObject.contentTypeId}/?expand=columns`;
        request.Method = 'GET';
        request.AuthToken = token;
        request.AuthScheme = 'Bearer';
        request.Send();
		
		let fieldNames = [];
		const jsonResponse = JSON.parse(request.ResponseText);
  		if (request.Status == 200 || request.Status == 201) {
			fieldsMapped = true;
			if (!jsonResponse.columns)
				deleteFile(true, `SharePoint error ${jsonResponse.error.message}`);
        	var columns = jsonResponse.columns;
			for (let column of columns)
				fieldNames.push(column.displayName);
			for (let el of mapping)
			{
				const displayName = el.field.sharepoint_field_name;
				const foundElement = columns.filter(function(columns){ return columns.displayName == displayName });
				
                if (foundElement[0])
				{
					el.field.sharepoint_field_name = foundElement[0].name;
					el.field['sharepoint_type'] = foundElement[0].type;
				}
                else    
				{
					Context.LogMessage(`Allowed field names: [${fieldNames.join(", ")}]`);
					const errorMessage = `No fields with name "${displayName}" were found for the content type "${fileObject.contentTypeName}". Export is aborting`;
					if (fileObject.existsInitially && file_uploading_policy === 'OVERWRITE')
						throw errorMessage
					else
						deleteFile(true, errorMessage);
				}
			}
			Context.LogMessage("SharePoint field names in mapping have been successfully changed");
		}
		else
			deleteFile(true, `SharePoint error ${jsonResponse.error.message}`);
	}
	
	function changeColumnsNameAccordingToSPNamesList() {
		const request = Context.CreateHttpRequest();
        request.Url = apiSitesUrl + siteObject.id + '/lists/' + fileObject.listId + '/?expand=columns';
        request.Method = 'GET';
        request.AuthToken = token;
        request.AuthScheme = 'Bearer';
        request.Send();

		let fieldNames = [];
        const jsonResponse = JSON.parse(request.ResponseText);
  		if (request.Status == 200 || request.Status == 201) {
        	var columns = jsonResponse.columns;
			for (let column of columns)
				fieldNames.push(column.displayName);
			for (let el of mapping)
			{
				const displayName = el.field.sharepoint_field_name;
				const foundElement = columns.filter(function(columns){ return columns.displayName == displayName });
                if (foundElement[0])
					el.field.sharepoint_field_name = foundElement[0].name;
                else  
				{
					Context.LogMessage(`Allowed field names: [${fieldNames.join(", ")}]`);
					throw `No fields with name "${displayName}" were found in the specified list`;
				}	 
			}
			Context.LogMessage("SharePoint field names in mapping have been successfully changed");
		}else
			throw `SharePoint error ${jsonResponse.error.message}`;
	}
    
	function createListRecord(document) {
        //creating json body for the POST request
		if (fileObject.contentTypeName)
			Context.LogMessage(`Creating a list record with '${fileObject.contentTypeName}' content type`);
		else
			Context.LogMessage(`Creating a list record with`);
		
        const json = fileObject.changesJson;
        //return if json is empty
        if (!json.contentType && !json.fields)
		{
			Context.LogMessage(`Mapping and content type are empty`);
			return;
		}
        const request = Context.CreateHttpRequest();
		request.ThrowExceptionOnFailed = false;
        request.Url = apiSitesUrl + siteObject.id + '/lists/' + fileObject.listId + '/items';
        request.Method = 'POST';
        request.SetStringContent(JSON.stringify(json), ContentEncoding.UTF8, "application/json");
        request.AuthToken = token;
        request.AuthScheme = 'Bearer';
        request.Send();

        const jsonResponse = JSON.parse(request.ResponseText);
  		if (request.Status > 199 && request.Status < 300) 
			return jsonResponse.webUrl;
		else
			throw `Error setting list columns; SharePont error: ${jsonResponse.error.message}`;
    }

    function createFieldsJson(document) {
		let json = {};
		for (let i = 0; i < mapping.length; i++) 
		{
			let sharePointField = mapping[i].field.sharepoint_field_name;
			let value, documentField;
			if (mapping[i].field.vantage_field_name)
			{
				const fieldName = mapping[i].field.vantage_field_name;
				documentField = document.GetField(fieldName);
				if (!documentField)
					throw `The field ${fieldName} was not found in the Vantage document`;
				
				if (documentField.DataType === DataType.Date)
				{
					if (documentField.Value)
						value = new Date(documentField.Value).toISOString().replace('Z', '');
					else
						value = null;
				}
				else if (documentField.DataType === DataType.Choice) 
				{
					let array = [];
					for (let child of documentField.Children)
						if (child.Value == true)
							array.push(child.FullName.split('/').pop());
					
					json[`${sharePointField}@odata.type`] = 'Collection(Edm.String)';
					value = array;
				}
				else if (documentField.DataType === DataType.None)
				{
					let format = 'json';
					if (mapping[i].field.table_format)
						format = mapping[i].field.table_format.toLowerCase();
					
					if (documentField.Instances.length > 0)
					{
						if (format === "csv"){
							let csvValue = '';
							let firstRow = true;
							for (let j = 0; j < documentField.Instances.length; j++) 
							{
								let instance = documentField.Instances[j];
								for (let child of instance.Children)
								{
									let cValue = "";
									if (child.Value)
										cValue = child.Value.toString();
									if (firstRow)
										csvValue += child.Name + ';'
									else
										csvValue+= cValue + ';';
								}
								if (firstRow)
								{
									firstRow = false;
									j--;
								}
								csvValue += '\n';
							}
							value = `${csvValue}`;
						}
						else if (format === "json"){
							let jsonArray = new Array();
							for (let instance of documentField.Instances) 
							{
								let jsonObject = {};
								for (let child of instance.Children)
								{
									
									let cValue = "";
									if (child.Value)
										cValue = child.Value.toString();
									jsonObject[child.Name] = cValue;
								}
								jsonArray.push(jsonObject);
							}
							value = JSON.stringify(jsonArray);
						}
					}
					else if (documentField.Children)
					{
						let jsonObject = {};
						for (let child of documentField.Children)
						{
							let cValue = "";
							if (child.Value)
								cValue = child.Value.toString();
							jsonObject[child.Name] = cValue;
						}
						value = JSON.stringify(jsonObject);
					}
					
					if (!value)
						value = '"Unknown value"';
				}
				else 
				{
					if(mapping[i].field.sharepoint_type === 'text' || mapping[i].field.sharepoint_type === 'note')
					{
						if (documentField.Instances.length > 0)
						{
							let arr = [];
							for (let instance of documentField.Instances)
								if (instance.Value)
									arr.push(instance.Value);
								else
									arr.push(instance.Text);
							value = `"${documentField.Name}": ${JSON.stringify(arr)}`;
						}
						else if (documentField.Value)
							value = documentField.Value.toString();
						else
							value = documentField.Text;
					}
					else
						value = documentField.Value;
				}
					
			}
			else
				value = mapping[i].field.value;
			
			json[sharePointField] = value;
		}
		return json;
	}


    function uploadTextFile() {
        const request = Context.CreateHttpRequest();
        request.Url = 'https://graph.microsoft.com/v1.0/drives/' + fileObject.driveId + '/root:' + fileObject.rootFolder + '/' + fileObject.fileName + ':/content';
        request.Method = 'PUT';
        request.AuthToken = fileObject.token;
        request.AuthScheme = 'Bearer';
        request.SetStringContent(fileObject.exportResult, ContentEncoding.UTF8, "text/plain");
        request.Send();

        const jsonResponse = JSON.parse(request.ResponseText);
 		if (request.Status == 201 || request.Status == 200) {
			Context.LogMessage(`File ${fileObject.fileName} uploaded successfully; link: ${jsonResponse.webUrl}`);
        	return jsonResponse.parentReference.path;
		} else {
            throw `SharePoint error: ${request.ResponseText}`;
        }
    }

    function uploadFile() {
        const request = Context.CreateHttpRequest();
		request.ThrowExceptionOnFailed = false;
		let conflictBehavior = 'replace';
		if (file_uploading_policy === 'CREATE_COPY')
			conflictBehavior = 'rename';
		else if (file_uploading_policy === 'OVERWRITE')
			deleteFile(false, "");
		else if (file_uploading_policy === 'DO_NOT_UPLOAD_IF_FILE_EXISTS')
			conflictBehavior = 'fail';
        request.Url = 'https://graph.microsoft.com/v1.0/drives/' + fileObject.driveId + '/root:' + fileObject.rootFolder + '/' + fileObject.fileName + ':/content?@microsoft.graph.conflictBehavior=' + conflictBehavior;
        
		request.Method = 'PUT';
        request.AuthToken = fileObject.token;
        request.AuthScheme = 'Bearer';
		if (fileObject.exportFormat === 'CUSTOM_JSON')
			request.SetStringContent(fileObject.exportResult, ContentEncoding.UTF8, "text/plain");
		else
			request.SetFileContent(fileObject.exportResult);
        request.Send();
 		if (request.Status == 201 || request.Status == 200) {
			const jsonResponse = JSON.parse(request.ResponseText);
			fileObject.fileName = jsonResponse.name;
			if (file_uploading_policy === 'CREATE_COPY')
			{
				if (fileObject.existsInitially)
					Context.LogMessage(`Copy of the file has been created, new file name is: '${fileObject.fileName}'; link: ${jsonResponse.webUrl}`);
				else
					Context.LogMessage(`File ${fileObject.fileName} uploaded successfully; link: ${jsonResponse.webUrl}`);
			}
			else if (file_uploading_policy === 'OVERWRITE')
				Context.LogMessage(`File ${fileObject.fileName} overwritten successfully; link: ${jsonResponse.webUrl}`);
			else
				Context.LogMessage(`File ${fileObject.fileName} uploaded successfully; link: ${jsonResponse.webUrl}`);
			fileObject.fileName = jsonResponse.name;
        	return jsonResponse.parentReference.path;
		} 
		else if (request.Status == 409)
		{
			Context.LogMessage(`WARNING: File ${fileObject.fileName} already exists, skipping it`);
			return 409;
		}
		else if (request.Status == 413)
			throw `File cannot be uploaded. Size of file '${fileObject.fileName}' is bigger that 4 MB.`;	
		else 
			throw `SharePoint error: ${request.ResponseText}`;
    }
	
	function deleteFile(interrupt, message) {
		if (export_scenario != 'UPLOAD_FILE')
			throw `${message}. Export is aborting`;
		
        const request = Context.CreateHttpRequest();
		request.ThrowExceptionOnFailed = false;
        request.Url = 'https://graph.microsoft.com/v1.0/drives/' + fileObject.driveId + '/root:' + fileObject.rootFolder + '/' + fileObject.fileName + '';
        request.Method = 'DELETE';
        request.AuthToken = fileObject.token;
        request.AuthScheme = 'Bearer';
        request.Send();
		
		if (fileObject.fileName)
			Context.LogMessage(`Error: ${fileObject.fileName} has been deleted`);
		if (interrupt)
			throw `${message}. Export is aborting`;
    }

    function getFileName(document, fileNameSchema) {
        let fileName = '';
        const parts = fileNameSchema.split('$(');

        let values = [];
        let delimiters = [];
        for (let i = 0; i < parts.length; i++) {
            let smallerParts = parts[i].split(')');
            if (smallerParts.length === 2) {
                if (smallerParts[0].includes('ExtractedData')) {
					const fieldName = smallerParts[0].replace('ExtractedData.', '');
                    const documentfield = document.GetField(fieldName);
                    if (!documentfield)
						throw `The field '${fieldName}' was not found in the Vantage document`;
					values.push(documentfield.Text);
                    delimiters.push(smallerParts[1]);
                } 
				else if (smallerParts[0].includes('DateTime')) {
					const time = new Date().toISOString().substring(0, 19).replace('T', ' ');
					values.push(replaceAll(':', '-', time));
					delimiters.push(smallerParts[1]);
				}
				else if (smallerParts[0].includes('Date')) {
					values.push(new Date().toISOString().substring(0, 10));
					delimiters.push(smallerParts[1]);
				}
				else if (smallerParts[0].includes('Time')) {
					const time = new Date().toISOString().split('T')[1].substring(0, 8);
					values.push(replaceAll(':', '-', time));
					delimiters.push(smallerParts[1]);
				}
				else if (smallerParts[0].includes('FullFileName')) {
					values.push(document.SourceFiles[0].FileName);
					delimiters.push(smallerParts[1]);
				}
				else if (smallerParts[0].includes('TransactionID')){
					values.push(Context.Transaction.Id);
					delimiters.push(smallerParts[1]);
				}
				else {
                    let value = getKeyWordValue(smallerParts[0]);
                    values.push(value);
                    delimiters.push(smallerParts[1]);
                }
            } else if (smallerParts.length === 1) {
                if (smallerParts[0]) {
                    let value = getKeyWordValue(smallerParts[0]);
                    values.push(value);
                    delimiters.push('');
                }
            }
        }

        for (let i = 0; i < values.length; i++) {
            if (i === (delimiters.length))
                fileName += values[i];
            else
                fileName += values[i] + delimiters[i];
        }
        return fileName;
    }

    function getKeyWordValue(keyWord) {
        if (keyWord === 'FullFileName') {
            return 'FileName'
        } else {
            return keyWord;
        }
    }

    function getToken() {
        const request = Context.CreateHttpRequest();
        request.Url = 'https://login.microsoftonline.com/' + tenant + '/oauth2/v2.0/token';
        request.Method = 'POST';
        const loginContent = 'client_id=' + client_id + '&scope=https://graph.microsoft.com/.default&client_secret=' + client_secret + '&grant_type=client_credentials';
        request.SetStringContent(loginContent, ContentEncoding.UTF8, "application/x-www-form-urlencoded");
		request.Send();

        let jsonResponse = 'ERROR'; 
		if (request.ResponseText)
			jsonResponse = JSON.parse(request.ResponseText);
		if (request.Status == 200) {
			Context.LogMessage(`Auth token requested successfully`);
        	return jsonResponse.access_token;
		}
		throw `SharePoint token request error: ${jsonResponse.error_description}`;
    }

	// Auxiliary connectors functions
	// Extracts the SharePoint site link from the SharePoint address
	function getSiteLink(sharePointLink) {
        let preparedLink;

        //replacing "shared part" with empty string
        if (sharePointLink.includes('/:f:/r')) {
            sharePointLink = sharePointUrl.replace('/:f:/r', '');
        } else if (sharePointLink.includes('/:w:/r')) {
            sharePointLink = sharePointUrl.replace('/:w:/r', '');
        } else if (sharePointLink.includes('/:l:/r')) {
            sharePointLink = sharePointUrl.replace('/:l:/r', '');
        }

        //remove https part from link
        preparedLink = sharePointLink.replace('https://', '').replace('/', ':');
        if (preparedLink.includes('?'))
            preparedLink = preparedLink.split('?')[0];
        if (preparedLink.includes('sites')) {
            return preparedLink.split('/')[0] + ':' + preparedLink.split('/')[1];
        } else if (preparedLink.includes('Lists'))
        {
            return preparedLink.split('/')[0] + ':' + "Lists";
        }else {
            let result = preparedLink.replace('/', ':');
            while (result.includes('/'))
                result = result.replace('/', ':');
            return result;
        }
    }

	// Requests the sine object using sire link
	function getSiteObject(siteLink) {
        const request = Context.CreateHttpRequest();
        request.Url = apiSitesUrl + siteLink;
        request.Method = 'GET';
        request.AuthToken = token;
        request.AuthScheme = 'Bearer';
        request.Send();

        const jsonResponse = JSON.parse(request.ResponseText);
		if (request.Status == 200) {
			return jsonResponse;
		}
		throw `SharePoint error: ${jsonResponse.error.message}`;
    }
	
	function getRootFolder(sharePointUrl, driveUrl) {
        const domainName = sharePointUrl.replace('https://', '').split('/')[0];
        let rootFolder;
        sharePointUrl = decodeURIComponent(sharePointUrl);
        driveUrl = decodeURIComponent(driveUrl);
        if (sharePointUrl.includes('/:f:/r')) {
            rootFolder = sharePointUrl.replace('/:f:/r', '').replace(driveUrl, '').split('?')[0];
        } else if (sharePointUrl.includes('sites')) {
            if (sharePointUrl.includes('&id='))
                rootFolder = 'https://' + domainName + sharePointUrl.split('&id=')[1].split('&')[0];
            else if (sharePointUrl.includes('?id='))
                rootFolder = 'https://' + domainName + sharePointUrl.split('?id=')[1].split('&')[0];
            else if (sharePointUrl.includes('Forms'))
                rootFolder = sharePointUrl.split('Forms')[0];
            rootFolder = rootFolder.replace(driveUrl, '');
        } else if(sharePointUrl.includes('RootFolder')){
            let fullPath = 'https://' + domainName + sharePointUrl.split('RootFolder')[1].split('&')[0].replace('=', '');
            rootFolder = fullPath.replace(driveUrl, '');
        }
		else{
			rootFolder = '';
		}
        return rootFolder;
    }
	
	function getDriveUrl(sharePointUrl, siteWebUrl) {
        sharePointUrl = sharePointUrl.replace('/:f:/r', '');
        const drivePath = sharePointUrl.replace(siteWebUrl, '').split('/')[1];
        return siteWebUrl + '/' + drivePath;
    }

	function getDriveIdBySiteId(siteId, driveUrl) {
        const request = Context.CreateHttpRequest();
        request.Url = apiSitesUrl + siteId + '/drives';
        request.Method = 'GET';
        request.AuthToken = token;
        request.AuthScheme = 'Bearer';
        request.Send();

        const jsonResponse = JSON.parse(request.ResponseText);
        if (request.Status == 200) {
            const propertyName = '@odata.nextLink';
            const nextLink = jsonResponse[propertyName];

            const values = jsonResponse.value;
            for (let i = 0; i < values.length; i++) {
                if (values[i].webUrl === driveUrl) {
                    return values[i].id;
                }
            }

            if (nextLink) {
                getDriveIdBySiteId(siteId, driveUrl);
            }
        }
        throw `SharePoint error: ${jsonResponse.error.message}`;
    }
	
	function replaceLast(find, replace, string) {
        let lastIndex = string.lastIndexOf(find);

        if (lastIndex === -1) {
            return string;
        }

        let beginString = string.substring(0, lastIndex);
        let endString = string.substring(lastIndex + find.length);

        return beginString + replace + endString;
    }

    function replaceAll(find, replace, string) {
        while (string.includes(find))
            string = string.replace(find, replace)
        return string;
    }

	function removeSpacesFromArray(array)
	{
		for (let i = 0; i < array.length; i++)
			array[i] = replaceAll(' ', '', array[i]);
	}
	// Validates if the uploading file exists on the SharePoint site
	function checkIfFileExists(driveId, rootFolder, fileName) {
        try {
            const request = Context.CreateHttpRequest();
			request.ThrowExceptionOnFailed = false;
            request.Url = 'https://graph.microsoft.com/v1.0/drives/' + fileObject.driveId + '/root:' + fileObject.rootFolder + '/' + fileObject.fileName;
            request.Method = 'GET';
            request.AuthToken = token;
            request.AuthScheme = 'Bearer';
            request.Send();

            if (request.Status > 199 && request.Status < 300) {
                return true;
            } else {
                return false;
            }
        } catch (Exception) {
            return false;
        }
    }

    return exporter;
}(config));

for (let i = 0; i < Context.Transaction.Documents.length; i++) {
	SharePointExporter.Export(Context.Transaction.Documents[i], (i + 1));
}

Context.LogMessage(`Exported ${exportedDocs} documents`);
// SIG // Begin signature block
// SIG // MIIm/QYJKoZIhvcNAQcCoIIm7jCCJuoCAQExDzANBglg
// SIG // hkgBZQMEAgEFADB3BgorBgEEAYI3AgEEoGkwZzAyBgor
// SIG // BgEEAYI3AgEeMCQCAQEEEBDgyQbOONQRoqMAEEvTUJAC
// SIG // AQACAQACAQACAQACAQAwMTANBglghkgBZQMEAgEFAAQg
// SIG // N60mf5oF7lfrc250yfSmYZyB0owNfmUqJnBa+nFKZ7mg
// SIG // ggoSMIIE0DCCA7igAwIBAgIBBzANBgkqhkiG9w0BAQsF
// SIG // ADCBgzELMAkGA1UEBhMCVVMxEDAOBgNVBAgTB0FyaXpv
// SIG // bmExEzARBgNVBAcTClNjb3R0c2RhbGUxGjAYBgNVBAoT
// SIG // EUdvRGFkZHkuY29tLCBJbmMuMTEwLwYDVQQDEyhHbyBE
// SIG // YWRkeSBSb290IENlcnRpZmljYXRlIEF1dGhvcml0eSAt
// SIG // IEcyMB4XDTExMDUwMzA3MDAwMFoXDTMxMDUwMzA3MDAw
// SIG // MFowgbQxCzAJBgNVBAYTAlVTMRAwDgYDVQQIEwdBcml6
// SIG // b25hMRMwEQYDVQQHEwpTY290dHNkYWxlMRowGAYDVQQK
// SIG // ExFHb0RhZGR5LmNvbSwgSW5jLjEtMCsGA1UECxMkaHR0
// SIG // cDovL2NlcnRzLmdvZGFkZHkuY29tL3JlcG9zaXRvcnkv
// SIG // MTMwMQYDVQQDEypHbyBEYWRkeSBTZWN1cmUgQ2VydGlm
// SIG // aWNhdGUgQXV0aG9yaXR5IC0gRzIwggEiMA0GCSqGSIb3
// SIG // DQEBAQUAA4IBDwAwggEKAoIBAQC54MsQ1K92vdSTYusw
// SIG // ZLiBCGzDBNliF44v/z5lz4/OYuY8UhzaFkVLVat4a2OD
// SIG // YpDOD2lsmcgaFItMzEUz6ojcnqOvK/6AYZ15V8TPLvQ/
// SIG // MDxdR/yaFrzDN5ZBUY4RS1T4KL7QjL7wMDge87Am+GZH
// SIG // Y23ecSZHjzhHU9FGHbTj3ADqRay9vHHZqm8A29vNMDp5
// SIG // T19MR/gd71vCxJ1gO7GyQ5HYpDNO6rPWJ0+tJYqlxvTV
// SIG // 0KaudAVkV4i1RFXULSo6Pvi4vekyCgKUZMQWOlDxSq7n
// SIG // eTOvDCAHf+jfBDnCaQJsY1L6d8EbyHSHyLmTGFBUNUtp
// SIG // Trw700kuH9zB0lL7AgMBAAGjggEaMIIBFjAPBgNVHRMB
// SIG // Af8EBTADAQH/MA4GA1UdDwEB/wQEAwIBBjAdBgNVHQ4E
// SIG // FgQUQMK9J47MNIMwojPX+2yz8LQsgM4wHwYDVR0jBBgw
// SIG // FoAUOpqFBxBnKLbv9r0FQW4gwZTaD94wNAYIKwYBBQUH
// SIG // AQEEKDAmMCQGCCsGAQUFBzABhhhodHRwOi8vb2NzcC5n
// SIG // b2RhZGR5LmNvbS8wNQYDVR0fBC4wLDAqoCigJoYkaHR0
// SIG // cDovL2NybC5nb2RhZGR5LmNvbS9nZHJvb3QtZzIuY3Js
// SIG // MEYGA1UdIAQ/MD0wOwYEVR0gADAzMDEGCCsGAQUFBwIB
// SIG // FiVodHRwczovL2NlcnRzLmdvZGFkZHkuY29tL3JlcG9z
// SIG // aXRvcnkvMA0GCSqGSIb3DQEBCwUAA4IBAQAIfmyTEMg4
// SIG // uJapkEv/oV9PBO9sPpyIBslQj6Zz91cxG7685C/b+LrT
// SIG // W+C05+Z5Yg4MotdqY3MxtfWoSKQ7CC2iXZDXtHwlTxFW
// SIG // MMS2RJ17LJ3lXubvDGGqv+QqG+6EnriDfcFDzkSnE3AN
// SIG // kR/0yBOtg2DZ2HKocyQetawiDsoXiWJYRBuriSUBAA/N
// SIG // xBti21G00w9RKpv0vHP8ds42pM3Z2Czqrpv1KrKQ0U11
// SIG // GIo/ikGQI31bS/6kA1ibRrLDYGCD+H1QQc7CoZDDu+8C
// SIG // L9IVVO5EFdkKrqeKM+2xLXY2JtwE65/3YR8V3Idv7kaW
// SIG // KK2hJn0KCacuBKONvPi8BDABMIIFOjCCBCKgAwIBAgII
// SIG // AiQj8H9P4OQwDQYJKoZIhvcNAQELBQAwgbQxCzAJBgNV
// SIG // BAYTAlVTMRAwDgYDVQQIEwdBcml6b25hMRMwEQYDVQQH
// SIG // EwpTY290dHNkYWxlMRowGAYDVQQKExFHb0RhZGR5LmNv
// SIG // bSwgSW5jLjEtMCsGA1UECxMkaHR0cDovL2NlcnRzLmdv
// SIG // ZGFkZHkuY29tL3JlcG9zaXRvcnkvMTMwMQYDVQQDEypH
// SIG // byBEYWRkeSBTZWN1cmUgQ2VydGlmaWNhdGUgQXV0aG9y
// SIG // aXR5IC0gRzIwHhcNMjAwOTMwMTgxNDA0WhcNMjMwOTMw
// SIG // MTgxNDA0WjB8MQswCQYDVQQGEwJVUzEXMBUGA1UECBMO
// SIG // Tm9ydGggQ2Fyb2xpbmExEjAQBgNVBAcTCUNoYXJsb3R0
// SIG // ZTEfMB0GA1UEChMWQUJCWVkgRGV2ZWxvcG1lbnQgSW5j
// SIG // LjEfMB0GA1UEAxMWQUJCWVkgRGV2ZWxvcG1lbnQgSW5j
// SIG // LjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEB
// SIG // AOQbacxe0+9WbPZM43JqM63zBTmDM7e4HWvnLeTSU0kJ
// SIG // ulmc49UQ/WfWSo/TGvlykUCOcVLmcWHmSgWHc7POACna
// SIG // bgMqW6m9NzXPtrmL5+nSKbPbDjAqHSnAi4WEa4/NRtoL
// SIG // iV92niXlVqIG/q1NbbwlJZQ8Yz0t6F+k40h2NBDT5Db0
// SIG // zcYiBRr46AnK9+LtqQ89P0urkssa1JNy6yOwlFT/T59X
// SIG // 8DeNWD9K9UcoNyBTQdz2M4G4htH2a8c5TyLEBNGr23Mp
// SIG // 6JuwqLE514r8gbZ494wzd1jlabhBrNsKWtgNc+yhD51s
// SIG // TiqhtanUDzQHlzgXPqjVBDkIxKRHirGMOFMCAwEAAaOC
// SIG // AYUwggGBMAwGA1UdEwEB/wQCMAAwEwYDVR0lBAwwCgYI
// SIG // KwYBBQUHAwMwDgYDVR0PAQH/BAQDAgeAMDUGA1UdHwQu
// SIG // MCwwKqAooCaGJGh0dHA6Ly9jcmwuZ29kYWRkeS5jb20v
// SIG // Z2RpZzJzNS02LmNybDBdBgNVHSAEVjBUMEgGC2CGSAGG
// SIG // /W0BBxcCMDkwNwYIKwYBBQUHAgEWK2h0dHA6Ly9jZXJ0
// SIG // aWZpY2F0ZXMuZ29kYWRkeS5jb20vcmVwb3NpdG9yeS8w
// SIG // CAYGZ4EMAQQBMHYGCCsGAQUFBwEBBGowaDAkBggrBgEF
// SIG // BQcwAYYYaHR0cDovL29jc3AuZ29kYWRkeS5jb20vMEAG
// SIG // CCsGAQUFBzAChjRodHRwOi8vY2VydGlmaWNhdGVzLmdv
// SIG // ZGFkZHkuY29tL3JlcG9zaXRvcnkvZ2RpZzIuY3J0MB8G
// SIG // A1UdIwQYMBaAFEDCvSeOzDSDMKIz1/tss/C0LIDOMB0G
// SIG // A1UdDgQWBBRfeF9cHCzr2yJMXPcl08RCo+y2izANBgkq
// SIG // hkiG9w0BAQsFAAOCAQEAHPM9prGEpgl2h271nBhyu3R3
// SIG // db/9G6ivg12bEc7c5H1IBR3peVx4tv3/QjDJaksNd90y
// SIG // IfqMVeB5ZrsiaJUBxM7fdMuJ+WlxpS1GY/ryiByJ3BAr
// SIG // IqkEqEAPuEcCdW6KsEmGb3dpm3cmRYJzI5oUpgrpkhFD
// SIG // 5GL5OtKBoq5N3gJAu/9BE+DW8mVgcJSiBW62SwqTsZSl
// SIG // iEPbNOdKYy9ynEXqjDUEaAT1bGZxXECyT7SjHdojbFP7
// SIG // 3XRUnsmBbK/FtkxHInkYY/nGsz78mXuYU52kwBoy7yJt
// SIG // YrFT5m6e/ol23TdvTnvs1DQ54Pfty49BZUhje9rM6+Yn
// SIG // dokFoem83jGCHEMwghw/AgEBMIHBMIG0MQswCQYDVQQG
// SIG // EwJVUzEQMA4GA1UECBMHQXJpem9uYTETMBEGA1UEBxMK
// SIG // U2NvdHRzZGFsZTEaMBgGA1UEChMRR29EYWRkeS5jb20s
// SIG // IEluYy4xLTArBgNVBAsTJGh0dHA6Ly9jZXJ0cy5nb2Rh
// SIG // ZGR5LmNvbS9yZXBvc2l0b3J5LzEzMDEGA1UEAxMqR28g
// SIG // RGFkZHkgU2VjdXJlIENlcnRpZmljYXRlIEF1dGhvcml0
// SIG // eSAtIEcyAggCJCPwf0/g5DANBglghkgBZQMEAgEFAKB8
// SIG // MBAGCisGAQQBgjcCAQwxAjAAMBkGCSqGSIb3DQEJAzEM
// SIG // BgorBgEEAYI3AgEEMBwGCisGAQQBgjcCAQsxDjAMBgor
// SIG // BgEEAYI3AgEVMC8GCSqGSIb3DQEJBDEiBCAziDjoZ1Bk
// SIG // Tgg4ZAKyZ3Zhk/wKd+y+/8B9GfKM0U98TDANBgkqhkiG
// SIG // 9w0BAQEFAASCAQCJrvsu1znY8RoBLE8pp32VFhK1ZoiK
// SIG // kB0/zq+vOzTa5GrqlOFfV4yJVQnygNBkLiGt+2PZK3Gv
// SIG // vAperIvnlcAfxhJLSCzvwVu/OwdMcDQceAUC7OgJLf6/
// SIG // CEMqyBp5CQR7qRdtFxvSjba+w2zzN8hICQ5tCJZ20X4Z
// SIG // 452mcq2B5XpiTjVYLkYTCJo5BpFBVMrNv/yVCl0zG6Zu
// SIG // jAJQU1t1xI7Cm1pMO4U5xYH9UJMoc9AIiW6Fx9J9LSjC
// SIG // KF12L251TJwlqqaol/wguvGsfHbthnfPQ8dXkLgW3EJe
// SIG // bN4Ndr6llCPBTbcmhc03OtB10ENuJL/vrgmSSddrbFnY
// SIG // aog0oYIZ1DCCGdAGCisGAQQBgjcDAwExghnAMIIZvAYJ
// SIG // KoZIhvcNAQcCoIIZrTCCGakCAQMxDTALBglghkgBZQME
// SIG // AgEwgdwGCyqGSIb3DQEJEAEEoIHMBIHJMIHGAgEBBgkr
// SIG // BgEEAaAyAgMwMTANBglghkgBZQMEAgEFAAQg3JH5OVEE
// SIG // HruM4YRDZZXxFgjtBm0eH8TWSErAbCFA+CUCFHkyWv3s
// SIG // ISotmruKwh+tRljp3AxhGA8yMDIyMDcxNTEzMTcyN1ow
// SIG // AwIBAaBXpFUwUzELMAkGA1UEBhMCQkUxGTAXBgNVBAoM
// SIG // EEdsb2JhbFNpZ24gbnYtc2ExKTAnBgNVBAMMIEdsb2Jh
// SIG // bHNpZ24gVFNBIGZvciBBZHZhbmNlZCAtIEc0oIIVZzCC
// SIG // BlgwggRAoAMCAQICEAHCnHr0eqYCWA6vMrEjsR0wDQYJ
// SIG // KoZIhvcNAQELBQAwWzELMAkGA1UEBhMCQkUxGTAXBgNV
// SIG // BAoTEEdsb2JhbFNpZ24gbnYtc2ExMTAvBgNVBAMTKEds
// SIG // b2JhbFNpZ24gVGltZXN0YW1waW5nIENBIC0gU0hBMzg0
// SIG // IC0gRzQwHhcNMjIwNDA2MDc0NDEyWhcNMzMwNTA4MDc0
// SIG // NDEyWjBTMQswCQYDVQQGEwJCRTEZMBcGA1UECgwQR2xv
// SIG // YmFsU2lnbiBudi1zYTEpMCcGA1UEAwwgR2xvYmFsc2ln
// SIG // biBUU0EgZm9yIEFkdmFuY2VkIC0gRzQwggGiMA0GCSqG
// SIG // SIb3DQEBAQUAA4IBjwAwggGKAoIBgQCj3qYhEhYSvCjg
// SIG // BPez1LDTAWiPU7YYWFtxoF7Y1kxz4FfwuQwH94e5KYP+
// SIG // 8NV1oUj/PbAcQLyCVWhIOgxG6z/DLJg0z8SYm3AbhhGN
// SIG // XhV8oQ3a1nd9r+x+nBTspb8pauuKKRr+Dp8suhZNWFpj
// SIG // cQzbLrRwCudGEELue0V8/mRFlK/g61CyUmcfcUM38eIY
// SIG // g0AQV5oV1/Lya56byVbbZ4MYePdlpAXM5hOFFP5fiWcB
// SIG // YfvaxoMo1o1O3TQsGAMBhEjdFngl4dZIaa1cNZYhHqDD
// SIG // TxMAF8vCXtySTQRiyXj13qexhAqedDqC3ICUtwFtq6g5
// SIG // nhpdwXwBBl2Qez5dSijKKRCxs1nPAbghMMfZtfSXLDau
// SIG // UsezMiNug6b51CT3VvhhdXRO8garIoTI/WTlXxWl3Cd0
// SIG // qtLQ6bRIeNeYzLsf+NZGw3V1+p5FxpV1awcHqETdVnYo
// SIG // zkpNAnlrT5Hi/Kyd67yKr3prbGQ0RvHMeBy8J/R1aKcz
// SIG // yToXfopORD6D870CAwEAAaOCAZ4wggGaMA4GA1UdDwEB
// SIG // /wQEAwIHgDAWBgNVHSUBAf8EDDAKBggrBgEFBQcDCDAd
// SIG // BgNVHQ4EFgQUSTtntVeimeZ0GXoMWSw+COoge4swTAYD
// SIG // VR0gBEUwQzBBBgkrBgEEAaAyAR4wNDAyBggrBgEFBQcC
// SIG // ARYmaHR0cHM6Ly93d3cuZ2xvYmFsc2lnbi5jb20vcmVw
// SIG // b3NpdG9yeS8wDAYDVR0TAQH/BAIwADCBkAYIKwYBBQUH
// SIG // AQEEgYMwgYAwOQYIKwYBBQUHMAGGLWh0dHA6Ly9vY3Nw
// SIG // Lmdsb2JhbHNpZ24uY29tL2NhL2dzdHNhY2FzaGEzODRn
// SIG // NDBDBggrBgEFBQcwAoY3aHR0cDovL3NlY3VyZS5nbG9i
// SIG // YWxzaWduLmNvbS9jYWNlcnQvZ3N0c2FjYXNoYTM4NGc0
// SIG // LmNydDAfBgNVHSMEGDAWgBTqFsZp5+PLV0U5M6TwQL7Q
// SIG // w71lljBBBgNVHR8EOjA4MDagNKAyhjBodHRwOi8vY3Js
// SIG // Lmdsb2JhbHNpZ24uY29tL2NhL2dzdHNhY2FzaGEzODRn
// SIG // NC5jcmwwDQYJKoZIhvcNAQELBQADggIBAAiIpOqFiDNY
// SIG // a378AFUi029DaW98/r5DhjcOn1PaUZ7QPOMbmddRURkP
// SIG // UZAO2+6aRs99CuDG7BWC6z6uzAP6We2PpUqiGT7CX/0W
// SIG // gzFWUihLX8RFg2HrlgwgMKJ9ReqWbbL8dLj9TGGUaqew
// SIG // 6qm/OI6YdnUHpvul3AtvdXpk6TDXkZBi0OHLGeToLyeI
// SIG // QIyH2z/bFbBIjeKNlYwn86xJh7B86cSl4NncvvYNFbje
// SIG // Y519liutpK6UYDfQSJmo270vTvQAj7f8SNq2EEDEPznb
// SIG // VXe9CzysNqBKmRTg0DEeidInCnBQ3a1vZPpvjRr2UPQW
// SIG // EzAMGM7YaELVVeNaX8CggbwZvESwY4p+wCseVW7nHR4T
// SIG // ZJlmZAmD6YHmPiv95HzsQ7ubbzVik2Sau1i4rwRuKLsK
// SIG // WOOFOSXU44sVcwE6HctdkfyeRS6HtfBGnJTDaK36DutH
// SIG // 2akl1ooK2J7vrKJepi6cWNG9Ub8SctARm0zPm1K/p+pK
// SIG // lCL82nSzRdSzCdZoREAqH4ps2uVpcQAS+Mnf6pipKmqP
// SIG // 1JrgH6yZ4ehdPnk8RaQOCoYUoBXlkiCB8oKO86rJnF8c
// SIG // SXT8IbUo4IEN/7d/mIPAYNMBxWbMYbzCpAsDNzaaMiXC
// SIG // xeaDlPzQeb+D07xTteP+z+FxPgXbYo8kve9TqvKeUww/
// SIG // fGvw6iU4X3KqMIIGWTCCBEGgAwIBAgINAewckkDe/S5A
// SIG // XXxHdDANBgkqhkiG9w0BAQwFADBMMSAwHgYDVQQLExdH
// SIG // bG9iYWxTaWduIFJvb3QgQ0EgLSBSNjETMBEGA1UEChMK
// SIG // R2xvYmFsU2lnbjETMBEGA1UEAxMKR2xvYmFsU2lnbjAe
// SIG // Fw0xODA2MjAwMDAwMDBaFw0zNDEyMTAwMDAwMDBaMFsx
// SIG // CzAJBgNVBAYTAkJFMRkwFwYDVQQKExBHbG9iYWxTaWdu
// SIG // IG52LXNhMTEwLwYDVQQDEyhHbG9iYWxTaWduIFRpbWVz
// SIG // dGFtcGluZyBDQSAtIFNIQTM4NCAtIEc0MIICIjANBgkq
// SIG // hkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA8ALiMCP64Bvh
// SIG // mnSzr3WDX6lHUsdhOmN8OSN5bXT8MeR0EhmW+s4nYluu
// SIG // B4on7lejxDXtszTHrMMM64BmbdEoSsEsu7lw8nKujPeZ
// SIG // Wl12rr9EqHxBJI6PusVP/zZBq6ct/XhOQ4j+kxkX2e4x
// SIG // z7yKO25qxIjw7pf23PMYoEuZHA6HpybhiMmg5ZninvSc
// SIG // TD9dW+y279Jlz0ULVD2xVFMHi5luuFSZiqgxkjvyen38
// SIG // DljfgWrhsGweZYIq1CHHlP5CljvxC7F/f0aYDoc9emXr
// SIG // 0VapLr37WD21hfpTmU1bdO1yS6INgjcZDNCr6lrB7w/V
// SIG // mbk/9E818ZwP0zcTUtklNO2W7/hn6gi+j0l6/5Cx1Pcp
// SIG // Fdf5DV3Wh0MedMRwKLSAe70qm7uE4Q6sbw25tfZtVv6K
// SIG // HQk+JA5nJsf8sg2glLCylMx75mf+pliy1NhBEsFV/W6R
// SIG // xbuxTAhLntRCBm8bGNU26mSuzv31BebiZtAOBSGssREG
// SIG // Ixnk+wU0ROoIrp1JZxGLguWtWoanZv0zAwHemSX5cW7p
// SIG // nF0CTGA8zwKPAf1y7pLxpxLeQhJN7Kkm5XcCrA5XDAnR
// SIG // YZ4miPzIsk3bZPBFn7rBP1Sj2HYClWxqjcoiXPYMBOMp
// SIG // +kuwHNM3dITZHWarNHOPHn18XpbWPRmwl+qMUJFtr1eG
// SIG // fhA3HWsaFN8CAwEAAaOCASkwggElMA4GA1UdDwEB/wQE
// SIG // AwIBhjASBgNVHRMBAf8ECDAGAQH/AgEAMB0GA1UdDgQW
// SIG // BBTqFsZp5+PLV0U5M6TwQL7Qw71lljAfBgNVHSMEGDAW
// SIG // gBSubAWjkxPioufi1xzWx/B/yGdToDA+BggrBgEFBQcB
// SIG // AQQyMDAwLgYIKwYBBQUHMAGGImh0dHA6Ly9vY3NwMi5n
// SIG // bG9iYWxzaWduLmNvbS9yb290cjYwNgYDVR0fBC8wLTAr
// SIG // oCmgJ4YlaHR0cDovL2NybC5nbG9iYWxzaWduLmNvbS9y
// SIG // b290LXI2LmNybDBHBgNVHSAEQDA+MDwGBFUdIAAwNDAy
// SIG // BggrBgEFBQcCARYmaHR0cHM6Ly93d3cuZ2xvYmFsc2ln
// SIG // bi5jb20vcmVwb3NpdG9yeS8wDQYJKoZIhvcNAQEMBQAD
// SIG // ggIBAH/iiNlXZytCX4GnCQu6xLsoGFbWTL/bGwdwxvsL
// SIG // Ca0AOmAzHznGFmsZQEklCB7km/fWpA2PHpbyhqIX3kG/
// SIG // T+G8q83uwCOMxoX+SxUk+RhE7B/CpKzQss/swlZlHb1/
// SIG // 9t6CyLefYdO1RkiYlwJnehaVSttixtCzAsw0SEVV3ezp
// SIG // Sp9eFO1yEHF2cNIPlvPqN1eUkRiv3I2ZOBlYwqmhfqJu
// SIG // FSbqtPl/KufnSGRpL9KaoXL29yRLdFp9coY1swJXH4uc
// SIG // /LusTN763lNMg/0SsbZJVU91naxvSsguarnKiMMSME6y
// SIG // CHOfXqHWmc7pfUuWLMwWaxjN5Fk3hgks4kXWss1ugnWl
// SIG // 2o0et1sviC49ffHykTAFnM57fKDFrK9RBvARxx0wxVFW
// SIG // YOh8lT0i49UKJFMnl4D6SIknLHniPOWbHuOqhIKJPsBK
// SIG // 9SH+YhDtHTD89szqSCd8i3VCf2vL86VrlR8EWDQKie2C
// SIG // UOTRe6jJ5r5IqitV2Y23JSAOG1Gg1GOqg+pscmFKyfpD
// SIG // xMZXxZ22PLCLsLkcMe+97xTYFEBsIB3CLegLxo1tjLZx
// SIG // 7VIh/j72n585Gq6s0i96ILH0rKod4i0UnfqWah3GPMrz
// SIG // 2Ry/U02kR1l8lcRDQfkl4iwQfoH5DZSnffK1CfXYYHJA
// SIG // UJUg1ENEvvqglecgWbZ4xqRqqiKbMIIFRzCCBC+gAwIB
// SIG // AgINAfJAQkDO/SLb6Wxx/DANBgkqhkiG9w0BAQwFADBM
// SIG // MSAwHgYDVQQLExdHbG9iYWxTaWduIFJvb3QgQ0EgLSBS
// SIG // MzETMBEGA1UEChMKR2xvYmFsU2lnbjETMBEGA1UEAxMK
// SIG // R2xvYmFsU2lnbjAeFw0xOTAyMjAwMDAwMDBaFw0yOTAz
// SIG // MTgxMDAwMDBaMEwxIDAeBgNVBAsTF0dsb2JhbFNpZ24g
// SIG // Um9vdCBDQSAtIFI2MRMwEQYDVQQKEwpHbG9iYWxTaWdu
// SIG // MRMwEQYDVQQDEwpHbG9iYWxTaWduMIICIjANBgkqhkiG
// SIG // 9w0BAQEFAAOCAg8AMIICCgKCAgEAlQfoc8pm+ewUyns8
// SIG // 9w0I8bRFCyyCtEjG61s8roO4QZIzFKRvf+kqzMawiGvF
// SIG // tonRxrL/FM5RFCHsSt0bWsbWh+5NOhUG7WRmC5KAykTe
// SIG // c5RO86eJf094YwjIElBtQmYvTbl5KE1SGooagLcZgQ5+
// SIG // xIq8ZEwhHENo1z08isWyZtWQmrcxBsW+4m0yBqYe+bnr
// SIG // qqO4v76CY1DQ8BiJ3+QPefXqoh8q0nAue+e8k7ttU+JI
// SIG // fIwQBzj/ZrJ3YX7g6ow8qrSk9vOVShIHbf2MsonP0KBh
// SIG // d8hYdLDUIzr3XTrKotudCd5dRC2Q8YHNV5L6frxQBGM0
// SIG // 32uTGL5rNrI55KwkNrfw77YcE1eTtt6y+OKFt3OiuDWq
// SIG // RfLgnTahb1SK8XJWbi6IxVFCRBWU7qPFOJabTk5aC0fz
// SIG // BjZJdzC8cTflpuwhCHX85mEWP3fV2ZGXhAps1AJNdMAU
// SIG // 7f05+4PyXhShBLAL6f7uj+FuC7IIs2FmCWqxBjplllnA
// SIG // 8DX9ydoojRoRh3CBCqiadR2eOoYFAJ7bgNYl+dwFnidZ
// SIG // THY5W+r5paHYgw/R/98wEfmFzzNI9cptZBQselhP00sI
// SIG // ScWVZBpjDnk99bOMylitnEJFeW4OhxlcVLFltr+Mm9wT
// SIG // 6Q1vuC7cZ27JixG1hBSKABlwg3mRl5HUGie/Nx4yB9gU
// SIG // YzwoTK8CAwEAAaOCASYwggEiMA4GA1UdDwEB/wQEAwIB
// SIG // BjAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQWBBSubAWj
// SIG // kxPioufi1xzWx/B/yGdToDAfBgNVHSMEGDAWgBSP8Et/
// SIG // qC5FJK5NUPpjmove4t0bvDA+BggrBgEFBQcBAQQyMDAw
// SIG // LgYIKwYBBQUHMAGGImh0dHA6Ly9vY3NwMi5nbG9iYWxz
// SIG // aWduLmNvbS9yb290cjMwNgYDVR0fBC8wLTAroCmgJ4Yl
// SIG // aHR0cDovL2NybC5nbG9iYWxzaWduLmNvbS9yb290LXIz
// SIG // LmNybDBHBgNVHSAEQDA+MDwGBFUdIAAwNDAyBggrBgEF
// SIG // BQcCARYmaHR0cHM6Ly93d3cuZ2xvYmFsc2lnbi5jb20v
// SIG // cmVwb3NpdG9yeS8wDQYJKoZIhvcNAQEMBQADggEBAEms
// SIG // XsWD81rLYSpNl0oVKZ/kFJCqCfnEep81GIoKMxVtcoci
// SIG // TkE/bQqeGK7b4l/8ldEsmBQ7jsHwNll5842Bz3T2GKTk
// SIG // 4WjP739lWULpylU5vNPFJu5xOPrXIQMPt07ZW2BqQ7R9
// SIG // CdBgYd2q7QBeTjIe4LJsnjyywruY05B2ammtGtyoidpY
// SIG // T9LCizJKzlT7OOk7Bwt1ChHbC3wlJ/GsJs8RU+bcxuJh
// SIG // NTL0zt2D4xk668Joo3IAyCQ8TrhTPLEXq+Y1LPnTQinm
// SIG // X2ADrEJhprFXajNC3zUxhso+NyvaxNok9U4S8ra5t0fq
// SIG // uyCtYRa3oDPjLYmnvLM8AX8jGoAJNOkwggNfMIICR6AD
// SIG // AgECAgsEAAAAAAEhWFMIojANBgkqhkiG9w0BAQsFADBM
// SIG // MSAwHgYDVQQLExdHbG9iYWxTaWduIFJvb3QgQ0EgLSBS
// SIG // MzETMBEGA1UEChMKR2xvYmFsU2lnbjETMBEGA1UEAxMK
// SIG // R2xvYmFsU2lnbjAeFw0wOTAzMTgxMDAwMDBaFw0yOTAz
// SIG // MTgxMDAwMDBaMEwxIDAeBgNVBAsTF0dsb2JhbFNpZ24g
// SIG // Um9vdCBDQSAtIFIzMRMwEQYDVQQKEwpHbG9iYWxTaWdu
// SIG // MRMwEQYDVQQDEwpHbG9iYWxTaWduMIIBIjANBgkqhkiG
// SIG // 9w0BAQEFAAOCAQ8AMIIBCgKCAQEAzCV2kHkGeCIW9cCD
// SIG // toTKKJ79BXYRxa2IcvxGAkPHsoqdBF8kyy5L4WCCRuFS
// SIG // qwyBR3Bs3WTR6/Usow+CPQwrrpfXthSGEHm7OxOAd4wI
// SIG // 4UnSamIvH176lmjfiSeVOJ8G1z7JyyZZDXPesMjpJg6D
// SIG // FcbvW4vSBGDKSaYo9mk79svIKJHlnYphVzesdBTcdOA6
// SIG // 7nIvLpz70Lu/9T0A4QYz6IIrrlOmOhZzjN1BDiA6wLSn
// SIG // oemyT5AuMmDpV8u5BJJoaOU4JmB1sp93/5EU764gSfyt
// SIG // QBVI0QIxYRleuJfvrXe3ZJp6v1/BE++bYvsNbOBUaRap
// SIG // A9pu6YOTcXbGaYWCFwIDAQABo0IwQDAOBgNVHQ8BAf8E
// SIG // BAMCAQYwDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4EFgQU
// SIG // j/BLf6guRSSuTVD6Y5qL3uLdG7wwDQYJKoZIhvcNAQEL
// SIG // BQADggEBAEtA28BQqv7IDO/3llRFSbuWAAlBrLMThoYo
// SIG // BzPKa+Z0uboALa6kCtP18fEPir9zZ0qDx0R7eOCvbmxv
// SIG // AymOMzlFw47kuVdsqvwSluxTxi3kJGy5lGP73FNoZ1Y+
// SIG // g7jPNSHDyWj+ztrCU6rMkIrp8F1GjJXdelgoGi8d3s0A
// SIG // N0GP7URt11Mol37zZwQeFdeKlrTT3kwnpEwbc3N29BeZ
// SIG // wh96DuMtCK0KHCz/PKtVDg+Rfjbrw1dJvuEuLXxgi8NB
// SIG // URMjnc73MmuUAaiZ5ywzHzo7JdKGQM47LIZ4yWEvFLru
// SIG // 21Vv34TuBQlNvSjYcs7TYlBlHuuSl4Mx2bO1ykdYP18x
// SIG // ggNJMIIDRQIBATBvMFsxCzAJBgNVBAYTAkJFMRkwFwYD
// SIG // VQQKExBHbG9iYWxTaWduIG52LXNhMTEwLwYDVQQDEyhH
// SIG // bG9iYWxTaWduIFRpbWVzdGFtcGluZyBDQSAtIFNIQTM4
// SIG // NCAtIEc0AhABwpx69HqmAlgOrzKxI7EdMAsGCWCGSAFl
// SIG // AwQCAaCCAS0wGgYJKoZIhvcNAQkDMQ0GCyqGSIb3DQEJ
// SIG // EAEEMCsGCSqGSIb3DQEJNDEeMBwwCwYJYIZIAWUDBAIB
// SIG // oQ0GCSqGSIb3DQEBCwUAMC8GCSqGSIb3DQEJBDEiBCCt
// SIG // V6ivR12q2NBIZYiO1EEyQI0DzQePdPSDkao+WQqNTTCB
// SIG // sAYLKoZIhvcNAQkQAi8xgaAwgZ0wgZowgZcEIK+AMe1u
// SIG // yzkUREiVvQsdDOsSlZTbXgwsbfa+crElQkfQMHMwX6Rd
// SIG // MFsxCzAJBgNVBAYTAkJFMRkwFwYDVQQKExBHbG9iYWxT
// SIG // aWduIG52LXNhMTEwLwYDVQQDEyhHbG9iYWxTaWduIFRp
// SIG // bWVzdGFtcGluZyBDQSAtIFNIQTM4NCAtIEc0AhABwpx6
// SIG // 9HqmAlgOrzKxI7EdMA0GCSqGSIb3DQEBCwUABIIBgF22
// SIG // GqU713XFnzl6scVuCwJqQyOcFXjoOiPkjXJxHVGJv7Os
// SIG // z1vas0XjSqsAhMyr8Vk8bqw5ljFD8k7nMQtu4DqVeWdX
// SIG // OOUPbRp0hxWDAiVssk7w44N4/AIXIfoFFgsjkY2Bl/FK
// SIG // wyf5ZoB3dA0MAzl8UZs3HuRIfiqv4lnxWs4U0x0dB44B
// SIG // kjlJ1tKqaesy+oTMF0pq9yfPJ3pBtMf2RI0YzCRplvJs
// SIG // C9/Q+yFpOjtEzdUZ3JpzwaC7/F+uo7aQJAyfOXUp5UBP
// SIG // QifJsUnbpgDWtQ6AcYKYJpNGVExpK3RGwjExqQBi6kOv
// SIG // evDrESNhzbuRzjjMiCX6HxF+IHj5mTnnZXaUR4npVAWf
// SIG // gFTyWSdmyR4JrZCbayAyDgJnIv6lhKAIaoMp8AeBvVej
// SIG // xX8/310mvgOa9Lp1a2sz7HhROAe6zJcLZt4jxQRBj2BB
// SIG // WCdCx7StcvPa4EZ+CwCaJFW5TqH2z7d93HocFBwcilFi
// SIG // tVHXmlLhf65g9hadvDmMkGRZPw==
// SIG // End signature block
